/*
 * Chess with reflexive rules---a player can
 * 'break the rules' a certain number of times,
 * after which they become the new rules for both sides.
 *
 * mark j. nelson, dec. 2006
 */

/* number of times rules can be broken (should be configurable?) */
int breaks = 1;

/* order to try inference in (e.g. pawn then bishop) */
int order[];

/* current game state */
int winner; /* game winner... -1 if still ongoing */
int xsel, ysel; /* -1 if no piece selected */
int turn; /* 0 for white, 1 for black */
int wbreak, bbreak; /* number of breaks left per side*/

int xcur, ycur;

/* use codes for pieces:
 * for white: 0 king, 1 rook, 2 bishop, 3 queen, 4 knight, 5 pawn
 * for black: 6 king, 7 rook, 8 bishop, 9 queen, 10 knight, 11 pawn
 * -1 for empty
 * (kind of a hack, since java enums are a pain) */
PImage images[];
int board[][]; /* row/col, with top/left = 0,0, black on top */
boolean highlight[][];

/* for the six types of pieces, which of the six types of
 * movement they can currently do?
 * movement[piecetype][movetype] indicates whether allowed */
boolean movement[][];

void setup()
{
  size(289,289); /* 35*8 + 9 */

  loadPieceImages();
  setupBoard();
  highlight = new boolean[8][8];
  for (int i = 0; i < 8; ++i)
    for (int j = 0; j < 8; ++j)
      highlight[i][j] = false;
  movement = new boolean[6][6];
  for (int i = 0; i < 6; ++i)
    for (int j = 0; j < 6; ++j)
      movement[i][j] = (i == j); /* initially, each piece does the normal thing */

  xsel = -1; ysel = -1;
  turn = 0;
  wbreak = breaks; bbreak = breaks;
  
  order = new int[6];
  order[0] = 5; order[1] = 0; order[2] = 4; order[3] = 2; order[4] = 1; order[5] = 3; /* queen not really needed */
  winner = -1;
  
  frameRate(30);
}

void draw()
{
  background(0);
  drawGrid();
  drawPieces();
  xcur = course(mouseY);
  ycur = course(mouseX);
  if (xsel == -1)
    pieceHighlight(xcur,ycur);
}

void mousePressed()
{
  int x = course(mouseY);
  int y = course(mouseX);
  if (x < 0 || x >= 8 || y < 0 || y >= 8)
    return;

  if (winner != -1)
    return;

  if (xsel == -1)
  {
    if (col(x,y) == turn)
    {
      xsel = x;
      ysel = y;
    }
    pieceHighlight(xsel, ysel);
  }
  else
  {
    if (highlight[x][y]) /* legal move */
    {
      move(xsel, ysel, x, y);
    }
    else if ((turn == 0 && wbreak > 0) || (turn == 1 && bbreak > 0)) /* see if we can infer a legal move */
    {
      for (int i = 0; i < 6; ++i)
      {
        highlightType(order[i], xsel, ysel);
        if (highlight[x][y])
        {
          movement[board[xsel][ysel] % 6][order[i]] = true;
          if (turn == 0) --wbreak; else --bbreak;
          move(xsel, ysel, x, y);
          break;
        }
      }
    }
    xsel = -1;
    ysel = -1;
    pieceHighlight(-1, -1);      
  }
}

/* assumes move is legal */
void move(int fromX, int fromY, int toX, int toY)
{
  if (board[toX][toY] == 0)
    winner = 1;
  if (board[toX][toY] == 6)
    winner = 0;
  board[toX][toY] = board[fromX][fromY];
  board[fromX][fromY] = -1;
  if (turn == 0)
    turn = 1;
  else
    turn = 0;
}

/* convert from pixels to squares */
int course(int pix)
{
  return pix / 36;
}

void pieceHighlight(int x, int y)
{
  if (!(x == -1 && y == -1) && (x < 0 || x >= 8 || y < 0 || y >= 8))
    return;
  
  for (int i = 0; i < 8; ++i)
    for (int j = 0; j < 8; ++j)
      highlight[i][j] = false;

  if (x == -1 || board[x][y] == -1)
    return;

  int type = board[x][y] % 6;

  for (int i = 0; i < 6; ++i)
    if (movement[type][i])
      highlightType(i, x, y);
}

/* 0 for white, 1 for black, -1 for empty, -2 for off board */
int col(int x, int y)
{
  if (x < 0 || x >= 8 || y < 0 || y >= 8)
    return -2;
  if (board[x][y] == -1)
    return -1;
  if (board[x][y] < 6)
    return 0;
  return 1;
}

void setHighlight(int x, int y)
{
  if (x >= 0 && x < 8 && y >= 0 && y < 8)
    highlight[x][y] = true;
}

void highlightType(int type, int x, int y)
{
  if (type == 0) /* king */
  {
    for (int i = -1; i <= 1; ++i)
      for (int j = -1; j <= 1; ++j)
        if (col(x+i,y+j) != col(x,y))
          setHighlight(x+i, y+j);
  }
  else if (type == 1) /* rook */
  {
    int i = x+1;
    while (col(i,y) == -1)
    {
      setHighlight(i,y);
      ++i;
    }
    if (col(i,y) != col(x,y)) 
      setHighlight(i,y);
    i = x-1;
    while (col(i,y) == -1)
    {
      setHighlight(i,y);
      --i;
    }
    if (col(i,y) != col(x,y)) 
      setHighlight(i,y);
    i = y+1;
    while (col(x,i) == -1)
    {
      setHighlight(x,i);
      ++i;
    }
    if (col(x,i) != col(x,y))
      setHighlight(x,i);
    i = y-1;
    while (col(x,i) == -1)
    {
      setHighlight(x,i);
      --i;
    }
    if (col(x,i) != col(x,y))
      setHighlight(x,i);
  }
  else if (type == 2) /* bishop */
  {
    int i = 1;
    while (col(x+i,y+i) == -1)
    {
      setHighlight(x+i,y+i);
      ++i;
    }
    if (col(x+i,y+i) != col(x,y))
      setHighlight(x+i,y+i);
    i = 1;
    while (col(x+i,y-i) == -1)
    {
      setHighlight(x+i,y-i);
      ++i;
    }
    if (col(x+i,y-i) != col(x,y))
      setHighlight(x+i,y-i);
    i = 1;
    while (col(x-i,y+i) == -1)
    {
      setHighlight(x-i, y+i);
      ++i;
    }
    if (col(x-i,y+i) != col(x,y))
      setHighlight(x-i,y+i);
    i = 1;
    while (col(x-i,y-i) == -1)
    {
      setHighlight(x-i,y-i);
      ++i;
    }
    if (col(x-i,y-i) != col(x,y))
      setHighlight(x-i,y-i);
  }
  else if (type == 3) /* queen */
  {
    highlightType(1, x, y);
    highlightType(2, x, y);
  }
  else if (type == 4) /* knight */
  {
    if (col(x+1,y+2) != col(x,y))
      setHighlight(x+1,y+2);
    if (col(x+2,y+1) != col(x,y))
      setHighlight(x+2,y+1);
    if (col(x+1,y-2) != col(x,y))
      setHighlight(x+1,y-2);
    if (col(x+2,y-1) != col(x,y))
      setHighlight(x+2,y-1);
    if (col(x-1,y+2) != col(x,y))
      setHighlight(x-1,y+2);
    if (col(x-2,y+1) != col(x,y))
      setHighlight(x-2,y+1);
    if (col(x-1,y-2) != col(x,y))
      setHighlight(x-1,y-2);
    if (col(x-2,y-1) != col(x,y))
      setHighlight(x-2,y-1);
  }
  else if (type == 5) /* pawn */
  {
    int dir = (col(x,y) == 0) ? -1 : 1; /* white goes "up" (neg numbers) */
    if (col(x+dir, y) == -1)
      setHighlight(x+dir, y);
    /* opening 2-space move, only for real pawns */
    if (((dir == -1 && x==6) || (dir == 1 && x==1)) && (board[x][y] == 5 || board[x][y] == 11) && col(x+2*dir,y) == -1)
      setHighlight(x+2*dir, y);
    /* captures */
    if (col(x+dir,y+1) != -1 && col(x+dir,y+1) != col(x,y))
      setHighlight(x+dir,y+1);
    if (col(x+dir,y-1) != -1 && col(x+dir,y-1) != col(x,y))
      setHighlight(x+dir,y-1);
  }
}

void drawGrid()
{
  stroke(0);
  for (int i = 0; i < 8; ++i)
  {
    for (int j = 0; j < 8; ++j)
    {
      if (winner != -1 && col(i,j) == winner)
        fill(100, 220, 100);
      else if (winner != -1)
        fill(255);
      else if (xsel == -1 && highlight[i][j])
        fill(220, 150, 150);
      else if (xsel != -1 && highlight[i][j])
        fill(200, 0, 0);
      else if (xsel == i && ysel == j)
        fill(50, 50, 220);
      else if (xsel == -1 && xcur == i && ycur == j)
        fill(150, 150, 220);
      else
        fill(255);
      rect(36*j, 36*i, 36, 36);
    }
  }
}

void drawPieces()
{
  for (int i = 0; i < 8; ++i)
    for (int j = 0; j < 8; ++j)
      if (board[i][j] >= 0)
        image(images[board[i][j]], 36*j+1, 36*i+1);
}

void loadPieceImages()
{
  images = new PImage[12];

  images[0] = loadImage("wk35.gif");
  images[1] = loadImage("wr35.gif");
  images[2] = loadImage("wb35.gif");
  images[3] = loadImage("wq35.gif");
  images[4] = loadImage("wn35.gif");
  images[5] = loadImage("wp35.gif");

  images[6] = loadImage("bk35.gif");
  images[7] = loadImage("br35.gif");
  images[8] = loadImage("bb35.gif");
  images[9] = loadImage("bq35.gif");
  images[10] = loadImage("bn35.gif");
  images[11] = loadImage("bp35.gif");
}

void setupBoard()
{
  board = new int[8][8];

  for (int i = 0; i < 8; ++i)
    for (int j = 0; j < 8; ++j)
      board[i][j] = -1;

  board[0][0] = 7;
  board[0][1] = 10;
  board[0][2] = 8;
  board[0][3] = 9;
  board[0][4] = 6;
  board[0][5] = 8;
  board[0][6] = 10;
  board[0][7] = 7;
  for (int i = 0; i < 8; ++i)
    board[1][i] = 11;

  board[7][0] = 1;
  board[7][1] = 4;
  board[7][2] = 2;
  board[7][3] = 3;
  board[7][4] = 0;
  board[7][5] = 2;
  board[7][6] = 4;
  board[7][7] = 1;
  for (int i = 0; i < 8; ++i)
    board[6][i] = 5;
}
